@@ -9,13 +9,37 @@ The majority of the "work" is defined in utils.py.
 import platform
 from functools import partial
 from operator import itemgetter
+from pathlib import PurePath
+from typing import (
+    Any,
+    Callable,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    TypeVar,
+    cast,
+)
 
 import natsort.compat.locale
 from natsort import utils
-from natsort.ns_enum import NS_DUMB, ns
+from natsort.ns_enum import NSType, NS_DUMB, ns
+from natsort.utils import NatsortInType, NatsortOutType
+
+# Common input and output types
+T = TypeVar("T")
+NatsortInTypeT = TypeVar("NatsortInTypeT", bound=NatsortInType)
+
+# The type that natsort_key returns
+NatsortKeyType = Callable[[NatsortInType], NatsortOutType]
+
+# Types for os_sorted
+OSSortKeyType = Callable[[NatsortInType], NatsortOutType]
 
 
-def decoder(encoding):
+def decoder(encoding: str) -> Callable[[Any], Any]:
     """
     Return a function that can be used to decode bytes to unicode.
 
@@ -56,7 +80,7 @@ def decoder(encoding):
     return partial(utils.do_decoding, encoding=encoding)
 
 
-def as_ascii(s):
+def as_ascii(s: Any) -> Any:
     """
     Function to decode an input with the ASCII codec, or return as-is.
 
@@ -79,7 +103,7 @@ def as_ascii(s):
     return utils.do_decoding(s, "ascii")
 
 
-def as_utf8(s):
+def as_utf8(s: Any) -> Any:
     """
     Function to decode an input with the UTF-8 codec, or return as-is.
 
@@ -102,7 +126,9 @@ def as_utf8(s):
     return utils.do_decoding(s, "utf-8")
 
 
-def natsort_keygen(key=None, alg=ns.DEFAULT):
+def natsort_keygen(
+    key: Optional[Callable[[Any], NatsortInType]] = None, alg: NSType = ns.DEFAULT
+) -> Callable[[Any], NatsortOutType]:
     """
     Generate a key to sort strings and numbers naturally.
 
@@ -212,7 +238,12 @@ natsort_keygen
 """
 
 
-def natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+def natsorted(
+    seq: Iterable[T],
+    key: Optional[Callable[[T], NatsortInType]] = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List[T]:
     """
     Sorts an iterable naturally.
 
@@ -257,11 +288,17 @@ def natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
         ['num2', 'num3', 'num5']
 
     """
-    key = natsort_keygen(key, alg)
-    return sorted(seq, reverse=reverse, key=key)
+    if alg & ns.PRESORT:
+        seq = sorted(seq, reverse=reverse, key=str)
+    return sorted(seq, reverse=reverse, key=natsort_keygen(key, alg))
 
 
-def humansorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+def humansorted(
+    seq: Iterable[T],
+    key: Optional[Callable[[T], NatsortInType]] = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List[T]:
     """
     Convenience function to properly sort non-numeric characters.
 
@@ -313,7 +350,12 @@ def humansorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
     return natsorted(seq, key, reverse, alg | ns.LOCALE)
 
 
-def realsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+def realsorted(
+    seq: Iterable[T],
+    key: Optional[Callable[[T], NatsortInType]] = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List[T]:
     """
     Convenience function to properly sort signed floats.
 
@@ -366,7 +408,12 @@ def realsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
     return natsorted(seq, key, reverse, alg | ns.REAL)
 
 
-def index_natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+def index_natsorted(
+    seq: Iterable[T],
+    key: Optional[Callable[[T], NatsortInType]] = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List[int]:
     """
     Determine the list of the indexes used to sort the input sequence.
 
@@ -422,20 +469,28 @@ def index_natsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
         ['baz', 'foo', 'bar']
 
     """
+    newkey: Callable[[Tuple[int, T]], NatsortInType]
     if key is None:
         newkey = itemgetter(1)
     else:
 
-        def newkey(x):
-            return key(itemgetter(1)(x))
+        def newkey(x: Tuple[int, T]) -> NatsortInType:
+            return cast(Callable[[T], NatsortInType], key)(itemgetter(1)(x))
 
     # Pair the index and sequence together, then sort by element
     index_seq_pair = [(x, y) for x, y in enumerate(seq)]
+    if alg & ns.PRESORT:
+        index_seq_pair.sort(reverse=reverse, key=lambda x: str(itemgetter(1)(x)))
     index_seq_pair.sort(reverse=reverse, key=natsort_keygen(newkey, alg))
     return [x for x, _ in index_seq_pair]
 
 
-def index_humansorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+def index_humansorted(
+    seq: Iterable[T],
+    key: Optional[Callable[[T], NatsortInType]] = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List[int]:
     """
     This is a wrapper around ``index_natsorted(seq, alg=ns.LOCALE)``.
 
@@ -484,7 +539,12 @@ def index_humansorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
     return index_natsorted(seq, key, reverse, alg | ns.LOCALE)
 
 
-def index_realsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
+def index_realsorted(
+    seq: Iterable[T],
+    key: Optional[Callable[[T], NatsortInType]] = None,
+    reverse: bool = False,
+    alg: NSType = ns.DEFAULT,
+) -> List[int]:
     """
     This is a wrapper around ``index_natsorted(seq, alg=ns.REAL)``.
 
@@ -529,8 +589,9 @@ def index_realsorted(seq, key=None, reverse=False, alg=ns.DEFAULT):
     return index_natsorted(seq, key, reverse, alg | ns.REAL)
 
 
-# noinspection PyShadowingBuiltins,PyUnresolvedReferences
-def order_by_index(seq, index, iter=False):
+def order_by_index(
+    seq: Sequence[Any], index: Iterable[int], iter: bool = False
+) -> Iterable[Any]:
     """
     Order a given sequence by an index sequence.
 
@@ -589,7 +650,7 @@ def order_by_index(seq, index, iter=False):
     return (seq[i] for i in index) if iter else [seq[i] for i in index]
 
 
-def numeric_regex_chooser(alg):
+def numeric_regex_chooser(alg: NSType) -> str:
     """
     Select an appropriate regex for the type of number of interest.
 
@@ -608,16 +669,19 @@ def numeric_regex_chooser(alg):
     return utils.regex_chooser(alg).pattern[1:-1]
 
 
-def _split_apply(v, key=None):
+def _split_apply(
+    v: Any, key: Optional[Callable[[T], NatsortInType]] = None
+) -> Iterator[str]:
     if key is not None:
         v = key(v)
-    return utils.path_splitter(str(v))
+    if not isinstance(v, (str, PurePath)):
+        v = str(v)
+    return utils.path_splitter(v)
 
 
 # Choose the implementation based on the host OS
 if platform.system() == "Windows":
-
-    from ctypes import wintypes, windll
+    from ctypes import wintypes, windll  # type: ignore
     from functools import cmp_to_key
 
     _windows_sort_cmp = windll.Shlwapi.StrCmpLogicalW
@@ -625,12 +689,15 @@ if platform.system() == "Windows":
     _windows_sort_cmp.restype = wintypes.INT
     _winsort_key = cmp_to_key(_windows_sort_cmp)
 
-    def os_sort_keygen(key=None):
-        return lambda x: tuple(map(_winsort_key, _split_apply(x, key)))
-
+    def os_sort_keygen(
+        key: Optional[Callable[[Any], NatsortInType]] = None
+    ) -> Callable[[Any], NatsortOutType]:
+        return cast(
+            Callable[[Any], NatsortOutType],
+            lambda x: tuple(map(_winsort_key, _split_apply(x, key))),
+        )
 
 else:
-
     # For UNIX-based platforms, ICU performs MUCH better than locale
     # at replicating the file explorer's sort order. We will use
     # ICU's ability to do basic natural sorting as it also better
@@ -645,14 +712,16 @@ else:
 
     except ImportError:
         # No ICU installed
-        def os_sort_keygen(key=None):
-            return natsort_keygen(
-                key=key, alg=ns.LOCALE | ns.PATH | ns.IGNORECASE
-            )
+        def os_sort_keygen(
+            key: Optional[Callable[[Any], NatsortInType]] = None
+        ) -> Callable[[Any], NatsortOutType]:
+            return natsort_keygen(key=key, alg=ns.LOCALE | ns.PATH | ns.IGNORECASE)
 
     else:
         # ICU installed
-        def os_sort_keygen(key=None):
+        def os_sort_keygen(
+            key: Optional[Callable[[Any], NatsortInType]] = None
+        ) -> Callable[[Any], NatsortOutType]:
             loc = natsort.compat.locale.get_icu_locale()
             collator = icu.Collator.createInstance(loc)
             collator.setAttribute(
@@ -699,14 +768,19 @@ os_sort_keygen
 """
 
 
-def os_sorted(seq, key=None, reverse=False):
+def os_sorted(
+    seq: Iterable[T],
+    key: Optional[Callable[[T], NatsortInType]] = None,
+    reverse: bool = False,
+    presort: bool = False,
+) -> List[T]:
     """
     Sort elements in the same order as your operating system's file browser
 
     .. warning::
 
         The resulting function will generate results that will be
-        differnt depending on your platform. This is intentional.
+        different depending on your platform. This is intentional.
 
     On Windows, this will sort with the same order as Windows Explorer.
 
@@ -722,7 +796,7 @@ def os_sorted(seq, key=None, reverse=False):
       special characters this will give correct results, but once
       special characters are added you should lower your expectations.
 
-    It is *strongly* reccommended to have :mod:`pyicu` installed on
+    It is *strongly* recommended to have :mod:`pyicu` installed on
     MacOS/Linux if you want correct sort results.
 
     It does *not* take into account if a path is a directory or a file
@@ -741,6 +815,10 @@ def os_sorted(seq, key=None, reverse=False):
         Return the list in reversed sorted order. The default is
         `False`.
 
+    presort : {{True, False}}, optional
+        Equivalent to adding ``ns.PRESORT``, see :class:`ns` for
+        documentation. The default is `False`.
+
     Returns
     -------
     out : list
@@ -756,4 +834,6 @@ def os_sorted(seq, key=None, reverse=False):
     This will implicitly coerce all inputs to str before collating.
 
     """
-    return sorted(seq, key=os_sort_keygen(key), reverse=reverse)
+    if presort:
+        seq = sorted(seq, reverse=reverse, key=str)
+    return sorted(seq, reverse=reverse, key=os_sort_keygen(key))
